from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)


from .managers.BroadcastManager import Broadcast
from .managers.CacheManager import Cache
from .managers.MailManager import Mail
from .managers.QueueManager import Queue
from .managers.SessionManager import Session
from .managers.UploadManager import Upload
from .environment import env
from .__version__ import (
    __title__,
    __description__,
    __url__,
    __version__,
    __author__,
    __author_email__,
    __licence__,
)

_file_source = "masonite"



"""Password Helper Module."""

import bcrypt


def password(password_string):
    """Bcrypt a string.

    Useful for storing passwords in a database.

    Arguments:
        pass {string} -- A string like a users plain text password to be bcrypted.

    Returns:
        string -- The encrypted string.
    """
    return bytes(
        bcrypt.hashpw(bytes(password_string, "utf-8"), bcrypt.gensalt())
    ).decode("utf-8")



"""Base queue driver."""

import pickle

import pendulum

from ...drivers import BaseDriver
from ...helpers import HasColoredCommands


class BaseQueueDriver(BaseDriver, HasColoredCommands):
    def add_to_failed_queue_table(self, payload, driver="amqp"):
        from config.database import DB as schema
        from config import queue

        if "amqp" in queue.DRIVERS:
            listening_channel = queue.DRIVERS["amqp"]["channel"]
        else:
            listening_channel = "default"

        if schema.get_schema_builder().has_table("failed_jobs"):
            schema.table("failed_jobs").insert(
                {
                    "driver": driver,
                    "channel": listening_channel,
                    "payload": pickle.dumps(payload),
                    "failed_at": pendulum.now(),
                }
            )

    def run_failed_jobs(self):
        from config.database import DB as schema

        try:
            self.success("Attempting to send failed jobs back to the queue ...")
            for job in schema.table("failed_jobs").get():
                payload = pickle.loads(job.payload)
                schema.table("failed_jobs").where("payload", job.payload).delete()
                self.push(
                    payload["obj"], args=payload["args"], callback=payload["callback"]
                )
        except Exception:
            self.danger("Could not get the failed_jobs table")

    def push(self, *objects, args=(), callback="handle", ran=1, channel=None):
        raise NotImplementedError

    def connect(self):
        return self

    def consume(self, channel, fair=False):
        raise NotImplementedError(
            "The {} driver does not implement consume".format(self.__class__.__name__)
        )

    def work(self):
        raise NotImplementedError(
            "The {} driver does not implement work".format(self.__class__.__name__)
        )



"""A RedirectionProvider Service Provider."""


from ..drivers import QueueAsyncDriver, QueueAmqpDriver, QueueDatabaseDriver
from ..managers import QueueManager
from ..provider import ServiceProvider
from .. import Queue
from ..helpers import config


class QueueProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind("QueueAsyncDriver", QueueAsyncDriver)
        self.app.bind("QueueAmqpDriver", QueueAmqpDriver)
        self.app.bind("QueueDatabaseDriver", QueueDatabaseDriver)
        self.app.bind("QueueManager", QueueManager)

    def boot(self, queue: QueueManager):
        self.app.bind("Queue", queue.driver(config("queue.driver")))
        self.app.swap(Queue, queue.driver(config("queue.driver")))



import json

from ..helpers import Dot
from ..request import Request


class MockRoute:
    def __init__(self, route, container, wsgi=None):
        self.route = route
        self.container = container
        self.wsgi = wsgi

    def assertIsNamed(self, name):
        assert self.route.named_route == name, "Route name is {}. Asserted {}".format(
            self.route.named_route, name
        )
        return self

    def assertIsNotNamed(self):
        assert self.route.named_route is None, "Route has a name: {}".format(
            self.route.named_route
        )
        return self

    def isNamed(self, name):
        return self.route.named_route == name

    def hasMiddleware(self, *middleware):
        return all(elem in self.route.list_middleware for elem in middleware)

    def hasController(self, controller):
        return self.route.controller == controller

    def assertHasController(self, controller):
        if isinstance(controller, str) and "@" in controller:
            controller, method = controller.split("@")
            assert (
                self.route.controller.__name__ == controller
            ), "Controller is {}. Asserted {}".format(
                self.route.controller.__name__, controller
            )
            assert (
                self.route.controller_method == method
            ), "Controller method is {}. Asserted {}".format(
                self.route.controller_method, method
            )

        return self

    def contains(self, value):
        return value in self.container.make("Response").decode("utf-8")

    def assertContains(self, value):
        assert self.contains(value), "Response does not contain {}".format(value)
        return self

    def assertNotFound(self):
        return self.assertIsStatus(404)

    def ok(self):
        return "200 OK" in self.container.make("Request").get_status_code()

    def canView(self):
        return self.ok()

    def get_string_response(self):
        response = self.container.make("Response")

        if isinstance(response, str):
            return response

        return response.decode("utf-8")

    def hasJson(self, key, value=""):

        response = json.loads(self.get_string_response())
        if isinstance(key, dict):
            for item_key, key_value in key.items():
                if not Dot().dot(item_key, response, False) == key_value:
                    return False
            return True
        return Dot().dot(key, response, False)

    def assertHasJson(self, key, value):
        response = json.loads(self.get_string_response())
        if isinstance(key, dict):
            for item_key, key_value in key.items():
                assert Dot().dot(item_key, response, False) == key_value
        else:
            assert (
                Dot().dot(key, response, False) == value
            ), "Key '{}' with the value of '{}' could not find a match in {}".format(
                key, value, response
            )
        return self

    def assertJsonContains(self, key, value):
        response = json.loads(self.get_string_response())
        if not isinstance(response, list):
            raise ValueError(
                "This method can only be used if the response is a list of elements."
            )

        found = False
        for element in response:
            if Dot().dot(key, element, False):
                assert Dot().dot(key, element, False)
                found = True

        if not found:
            raise AssertionError(
                "Could not find a key of: {} that had the value of {}".format(
                    key, value
                )
            )
        return self

    def count(self, amount):
        return len(json.loads(self.get_string_response())) == amount

    def assertCount(self, amount):
        response_amount = len(json.loads(self.get_string_response()))
        assert (
            response_amount == amount
        ), "Response has an count of {}. Asserted {}".format(response_amount, amount)
        return self

    def amount(self, amount):
        return self.count(amount)

    def hasAmount(self, key, amount):
        response = json.loads(self.get_string_response())
        try:
            return len(response[key]) == amount
        except TypeError:
            raise TypeError(
                "The json response key of: {} is not iterable but has the value of {}".format(
                    key, response[key]
                )
            )

    def assertHasAmount(self, key, amount):
        response = json.loads(self.get_string_response())
        try:
            assert len(response[key]) == amount, "{} is not equal to {}".format(
                len(response[key]), amount
            )
        except TypeError:
            raise TypeError(
                "The json response key of: {} is not iterable but has the value of {}".format(
                    key, response[key]
                )
            )

        return self

    def assertNotHasAmount(self, key, amount):
        response = json.loads(self.get_string_response())
        try:
            assert (
                not len(response[key]) == amount
            ), "{} is equal to {} but should not be".format(len(response[key]), amount)
        except TypeError:
            raise TypeError(
                "The json response key of: {} is not iterable but has the value of {}".format(
                    key, response[key]
                )
            )

        return self

    def user(self, obj):
        self._user = obj
        self.container.on_resolve(Request, self._bind_user_to_request)
        return self

    def isPost(self):
        return "POST" in self.route.method_type

    def isGet(self):
        return "GET" in self.route.method_type

    def isPut(self):
        return "PUT" in self.route.method_type

    def isPatch(self):
        return "PATCH" in self.route.method_type

    def isDelete(self):
        return "DELETE" in self.route.method_type

    def on_bind(self, obj, method):
        self.container.on_bind(obj, method)
        return self

    def hasSession(self, key):
        return self.container.make("Session").has(key)

    def assertParameterIs(self, key, value):
        request = self.container.make("Request")
        if key not in request.url_params:
            raise AssertionError(
                "Request class does not have the '{}' url parameter".format(key)
            )

        if request.param(key) != value:
            raise AssertionError(
                "parameter {} is equal to {} of type {}, not {} of type {}".format(
                    key,
                    request.param(key),
                    type(request.param(key)),
                    value,
                    type(value),
                )
            )

    def assertIsStatus(self, status):
        request = self.container.make("Request")
        assert request.is_status(status), AssertionError(
            "{} is not equal to {}".format(request.get_status_code(), status)
        )
        if not request.is_status(status):
            raise AssertionError(
                "{} is not equal to {}".format(request.get_status_code(), status)
            )

        return self

    def assertHasHeader(self, key):
        request = self.container.make("Request")
        assert request.header(key), "Header '{}' does not exist".format(key)
        return self

    def assertNotHasHeader(self, key):
        request = self.container.make("Request")
        assert not request.header(
            key
        ), "Header '{}' exists but asserting it should not".format(key)
        return self

    def assertHeaderIs(self, key, value):
        request = self.container.make("Request")
        assert str(request.header(key)) == str(value), AssertionError(
            "{} is not equal to {}".format(request.header(key), value)
        )

        return self

    def assertPathIs(self, url):
        path = self.container.make("Request").path
        assert path == url, "Asserting the path is '{}' but it is '{}'".format(
            url, path
        )
        return True

    def session(self, key):
        return self.container.make("Session").get(key)

    def on_make(self, obj, method):
        self.container.on_make(obj, method)
        return self

    def on_resolve(self, obj, method):
        self.container.on_resolve(obj, method)
        return self

    def _bind_user_to_request(self, request, container):
        request.set_user(self._user)
        return self

    def headerIs(self, key, value):
        request = self.container.make("Request")
        assertion = request.header(key) == value
        if not assertion:
            raise AssertionError(
                "header {} does not equal {}".format(request.header(key), value)
            )
        return assertion

    def parameterIs(self, key, value):
        request = self.container.make("Request")
        assertion = request.param(key) == value
        if not assertion:
            raise AssertionError(
                "parameter {} is equal to {} of type {}, not {} of type {}".format(
                    key,
                    request.param(key),
                    type(request.param(key)),
                    value,
                    type(value),
                )
            )
        return assertion

    @property
    def request(self):
        return self.container.make("Request")

    @property
    def response(self):
        """Gets the string response from the container. This isinstance check here
        is to support Python 3.5. Once python3.5 goes away we can can remove this check.

        @required for 3.5

        Returns:
            string
        """
        response = self.get_string_response()
        if isinstance(response, str):
            return response

        return response.decode("utf-8")

    def asDictionary(self):
        try:
            return json.loads(self.response)
        except ValueError:
            raise ValueError("The response was not json serializable")



"""Remove Preset"""
import os
import shutil
from ..presets import Preset


class Remove(Preset):
    """Removes any defined Preset"""

    def install(self):
        """Install the preset"""
        self.update_packages()
        self.update_bootstrapping()
        self.update_webpack_configuration()
        self.remove_node_modules()
        if os.path.exists(os.path.realpath("resources/sass/_variables.scss")):
            os.remove("resources/sass/_variables.scss")
        shutil.rmtree("resources/js/components", ignore_errors=True)
        shutil.rmtree("public/css", ignore_errors=True)
        shutil.rmtree("public/js", ignore_errors=True)

    def update_package_array(self, packages={}):
        """Updates the packages array to remove React, VueJS, and Bootstrap packages"""
        packages_to_remove = [
            "bootstrap",
            "jquery",
            "popper.js",
            "vue",
            "vue-template-compiler",
            "@babel/preset-react",
            "react",
            "react-dom",
        ]
        for package in packages_to_remove:
            packages.pop(package, None)

        return packages

    def update_webpack_configuration(self):
        """Copy webpack.mix.js file into application"""
        shutil.copyfile(
            os.path.dirname(__file__) + "/remove-stubs/webpack.mix.js", "webpack.mix.js"
        )

    def update_bootstrapping(self):
        """Copies template app.js file into application"""
        for directory in ["resources/sass", "resources/js"]:
            if not os.path.exists(os.path.realpath(directory)):
                os.makedirs(os.path.realpath(directory))
        with open("resources/sass/app.scss", "w") as f:
            f.write("")
        shutil.copyfile(
            os.path.dirname(__file__) + "/remove-stubs/app.js", "resources/js/app.js"
        )
        shutil.copyfile(
            os.path.dirname(__file__) + "/remove-stubs/bootstrap.js",
            "resources/js/bootstrap.js",
        )



"""Migrate Refresh Command."""

from cleo import Command


class MigrateRefreshCommand(Command):
    """
    Migrate refresh.

    migrate:refresh
        {--s|seed : Seed the database}
        {--c|connection=default : The connection you want to run migrations on}
    """

    def handle(self):
        self.call("migrate:reset", [("--connection", self.option("connection"))])
        self.call("migrate", [("--connection", self.option("connection"))])

        if self.option("seed"):
            self.call("seed:run")



"""Vue Preset"""
import os
import shutil
from ..presets import Preset


class Vue(Preset):
    """
    Configure the front-end scaffolding for the application to use VueJS

    Will also remove React as React and Vue are a bit mutally exclusive
    """

    def install(self):
        """Install the preset"""
        self.ensure_component_directory_exists()
        self.update_packages()
        self.update_webpack_configuration()
        self.update_bootstrapping()
        self.update_component()
        self.create_scss_file()
        self.remove_node_modules()

    def update_package_array(self, packages={}):
        """
        Updates the packages array to include VueJS specific packages
        but also remove React ones
        """
        for package in ["@babel/preset-react", "react", "react-dom"]:
            packages.pop(package, None)

        packages["vue"] = "^2.5.17"
        return packages

    def update_webpack_configuration(self):
        """Copy webpack.mix.js file into application"""
        shutil.copyfile(
            os.path.dirname(__file__) + "/vue-stubs/webpack.mix.js", "webpack.mix.js"
        )

    def update_component(self):
        """
        Copy example VueJS component into application
        (delete example React component if it exists)
        """
        vue_component = "resources/js/components/Example.js"
        if os.path.exists(os.path.realpath(vue_component)):
            os.remove(vue_component)
        shutil.copyfile(
            os.path.dirname(__file__) + "/vue-stubs/ExampleComponent.vue",
            "resources/js/components/ExampleComponent.vue",
        )

    def update_bootstrapping(self):
        """Copies template app.js and bootstrap.js into application"""
        shutil.copyfile(
            os.path.dirname(__file__) + "/vue-stubs/app.js", "resources/js/app.js"
        )
        shutil.copyfile(
            os.path.dirname(__file__) + "/shared-stubs/bootstrap.js",
            "resources/js/bootstrap.js",
        )



"""Session Cookie Module."""

import json
from email import message

from ...contracts import SessionContract
from ...drivers import BaseDriver
from ...helpers import config
from ...request import Request


class SessionCookieDriver(SessionContract, BaseDriver):
    """Cookie Session Driver."""

    def __init__(self, request: Request):
        """Cookie Session Constructor.

        Arguments:
            Environ {dict} -- The WSGI environment
            Request {masonite.request.Request} -- The Request class.
        """
        self.request = request

    def get(self, key):
        """Get a value from the session.

        Arguments:
            key {string} -- The key to get from the session.

        Returns:
            string|None - Returns None if a value does not exist.
        """
        cookie = self.request.get_cookie("s_{0}".format(key))
        if cookie:
            return self._get_serialization_value(cookie)

        cookie = self.request.get_cookie("f_{0}".format(key))
        if cookie:
            return self._get_serialization_value(cookie)

        return None

    def set(self, key, value):
        """Set a vlue in the session.

        Arguments:
            key {string} -- The key to set as the session key.
            value {string} -- The value to set in the session.
        """
        if isinstance(value, dict):
            value = json.dumps(value)

        self.request.cookie("s_{0}".format(key), value)

    def has(self, key):
        """Check if a key exists in the session.

        Arguments:
            key {string} -- The key to check for in the session.

        Returns:
            bool
        """
        if self.get(key):
            return True
        return False

    def all(self, flash_only=False):
        """Get all session data.

        Returns:
            dict
        """
        return self.__collect_data(flash_only=flash_only)

    def delete(self, key):
        """Delete a value in the session by it's key.

        Arguments:
            key {string} -- The key to find in the session.

        Returns:
            bool -- If the key was deleted or not
        """
        self.__collect_data()

        if self.request.get_cookie("s_{}".format(key)):
            self.request.delete_cookie("s_{}".format(key))
            return True

        return False

    def __collect_data(self, flash_only=False):
        """Collect data from session and flash data.

        Returns:
            dict
        """
        cookies = {}
        if (
            "HTTP_COOKIE" in self.request.environ
            and self.request.environ["HTTP_COOKIE"]
        ):
            cookies_original = self.request.environ["HTTP_COOKIE"].split(";")
            for cookie in cookies_original:
                if flash_only:
                    if cookie.strip().startswith("f_"):
                        data = cookie.split("=", 1)
                        cookie_name = (
                            data[0].replace("s_", "").replace("f_", "").strip()
                        )
                        cookies.update({cookie_name: self.get(cookie_name)})
                else:
                    if cookie.strip().startswith("s_") or cookie.strip().startswith(
                        "f_"
                    ):
                        data = cookie.split("=", 1)
                        cookie_name = (
                            data[0].replace("s_", "").replace("f_", "").strip()
                        )
                        cookies.update({cookie_name: self.get(cookie_name)})
        return cookies

    def flash(self, key, value):
        """Add temporary data to the session.

        Arguments:
            key {string} -- The key to set as the session key.
            value {string} -- The value to set in the session.
        """
        if isinstance(value, (dict, list)):
            value = json.dumps(value)

        self.request.cookie(
            "f_{0}".format(key),
            value,
            expires=config("session.drivers.cookie.flash_expires", "2 seconds"),
        )

    def get_error_messages(self):
        """Should get and delete the flashed messages

        Arguments:
            key {string} -- The key to set as the session key.
            value {string} -- The value to set in the session.
        """
        only_messages = []
        messages = self.all(flash_only=True).get("errors", {}).items()
        for key, messages in messages:
            for error_message in messages:
                only_messages.append(error_message)
        self.reset(flash_only=True)
        return only_messages

    def get_flashed_messages(self, key, value):
        """Should get and delete the flashed messages

        Arguments:
            key {string} -- The key to set as the session key.
            value {string} -- The value to set in the session.
        """
        messages = self.all(flash_only=True)
        self.reset(flash_only=True)
        return messages

    def reset(self, flash_only=False):
        """Delete all session data.

        Keyword Arguments:
            flash_only {bool} -- If only flash data should be deleted. (default: {False})
        """
        cookies = self.__collect_data()
        for cookie in cookies:
            if flash_only:
                self.request.delete_cookie("f_{0}".format(cookie))
                continue

            self.request.delete_cookie("s_{0}".format(cookie))

    def helper(self):
        """Use to create builtin helper function."""
        return self

    def _get_serialization_value(self, value):
        try:
            return json.loads(value)
        except ValueError:
            return value



"""A RedirectionProvider Service Provider."""

from ..drivers import BroadcastAblyDriver, BroadcastPusherDriver
from ..managers import BroadcastManager
from ..provider import ServiceProvider
from .. import Broadcast
from ..helpers import config


class BroadcastProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind("BroadcastPusherDriver", BroadcastPusherDriver)
        self.app.bind("BroadcastAblyDriver", BroadcastAblyDriver)
        self.app.bind("BroadcastManager", BroadcastManager(self.app))

    def boot(self):
        self.app.bind(
            "Broadcast",
            self.app.make("BroadcastManager").driver(config("broadcast.driver")),
        )
        self.app.swap(
            Broadcast,
            self.app.make("BroadcastManager").driver(config("broadcast.driver")),
        )



"""New Preset System Command."""
from cleo import Command
from ..commands.presets.React import React
from ..commands.presets.Vue import Vue
from ..commands.presets.Bootstrap import Bootstrap
from ..commands.presets.Remove import Remove


class PresetCommand(Command):
    """
    Swap the front-end scaffolding for the application

    preset
        {name : Name of the preset}
    """

    def handle(self):
        self.info("Scaffolding Application ...")
        preset_name = self.argument("name")
        if preset_name not in ["react", "vue", "remove", "bootstrap"]:
            raise ValueError("Invalid preset")
        return getattr(self, preset_name)()

    def remove(self):
        """Removes frontend scaffolding"""
        Remove().install()
        self.info("Frontend scaffolding removed successfully.")

    def react(self):
        """Add React frontend while also removing Vue (if it was previously selected)"""
        React().install()
        self.info("React scaffolding installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )

    def vue(self):
        """Add Vue frontend while also removing React (if it was previously selected)"""
        Vue().install()
        self.info("Vue scaffolding installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )

    def bootstrap(self):
        """Add Bootstrap Sass scafolding"""
        Bootstrap().install()
        self.info("Bootstrap scaffolding installed successfully.")
        self.comment(
            'Please run "npm install && npm run dev" to compile your fresh scaffolding.'
        )



"""Metadate for Masonite package."""
__title__ = "masonite"
__description__ = "The core for the Masonite framework"
__url__ = "https://github.com/MasoniteFramework/masonite"
__version__ = "2.3.24"
__author__ = "Joseph Mancuso"
__author_email__ = "joe@masoniteproject.com"
__licence__ = "MIT"



"""Authentication Class."""


class Auth:
    """Facade class for the Guard class"""

    @staticmethod
    def routes():
        from ..routes import Get, Post

        return [
            Get("/login", "auth.LoginController@show").name("login"),
            Get("/logout", "auth.LoginController@logout").name("logout"),
            Post("/login", "auth.LoginController@store"),
            Get("/register", "auth.RegisterController@show").name("register"),
            Post("/register", "auth.RegisterController@store"),
            Get("/home", "auth.HomeController@show").name("home"),
            Get("/email/verify", "auth.ConfirmController@verify_show").name("verify"),
            Get("/email/verify/send", "auth.ConfirmController@send_verify_email"),
            Get("/email/verify/@id:signed", "auth.ConfirmController@confirm_email"),
            Get("/password", "auth.PasswordController@forget").name("forgot.password"),
            Post("/password", "auth.PasswordController@send"),
            Get("/password/@token/reset", "auth.PasswordController@reset").name(
                "password.reset"
            ),
            Post("/password/@token/reset", "auth.PasswordController@update"),
        ]



"""New Job Command."""
from ..commands import BaseScaffoldCommand


class JobCommand(BaseScaffoldCommand):
    """
    Creates a new Job.

    job
        {name : Name of the job you want to create}
    """

    scaffold_name = "Job"
    template = "/masonite/snippets/scaffold/job"
    base_directory = "app/jobs/"
    postfix = "Job"



"""A QueueTableCommand Command"""


from cleo import Command

from ..helpers.filesystem import copy_migration


class QueueTableCommand(Command):
    """
    Create migration files for the queue feature

    queue:table
        {--failed : Created the queue failed table}
        {--jobs : Created the queue failed table}
    """

    def handle(self):
        if self.option("failed"):
            copy_migration("masonite/snippets/migrations/create_failed_jobs_table.py")
        if self.option("jobs"):
            copy_migration("masonite/snippets/migrations/create_queue_jobs_table.py")
        self.info("Migration created successfully")



"""Broadcast Manager Module."""

from ..contracts import BroadcastManagerContract
from .Manager import Manager


class BroadcastManager(Manager, BroadcastManagerContract):
    """Manages all broadcast drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = "broadcast"
    driver_prefix = "Broadcast"


class Broadcast:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""A Mail Service Provider."""

from ..drivers import (
    MailMailgunDriver,
    MailSmtpDriver,
    MailLogDriver,
    MailTerminalDriver,
)
from ..managers import MailManager
from ..provider import ServiceProvider
from .. import Mail
from ..helpers import config


class MailProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind("MailSmtpDriver", MailSmtpDriver)
        self.app.bind("MailMailgunDriver", MailMailgunDriver)
        self.app.bind("MailLogDriver", MailLogDriver)
        self.app.bind("MailTerminalDriver", MailTerminalDriver)
        self.app.bind("MailManager", MailManager(self.app))

    def boot(self, manager: MailManager):
        self.app.bind("Mail", manager.driver(config("mail.driver")))
        self.app.swap(Mail, manager.driver(config("mail.driver")))



"""Publish Service Providers"""
from cleo import Command


class PublishCommand(Command):
    """
    Publishes a Service Provider

    publish
        {name : Name of the Service Provider you want to publish}
        {--t|tag=None : The tag of the provider you want to publish}
    """

    def handle(self):
        from wsgi import container

        for provider in container.make("Providers"):
            if provider.__class__.__name__ == self.argument("name"):
                if self.option("tag") != "None":
                    provider.publish(tag=self.option("tag"))
                    provider.publish_migrations(tag=self.option("tag"))

                provider.publish()
                provider.publish_migrations()

                return

        raise ValueError("Could not find the {} provider".format(self.argument("name")))



"""A ModelDocstringCommand Command."""

from cleo import Command


class ModelDocstringCommand(Command):
    """
    Generate a model docstring based on a table definition

    model:docstring
        {table : Name of the table to generate the docstring for}
        {--c|connection=default : The connection to use}
    """

    def handle(self):
        from config.database import DB

        if self.option("connection") == "default":
            conn = DB.get_schema_manager().list_table_columns(self.argument("table"))
        else:
            conn = (
                DB.connection(self.option("connection"))
                .get_schema_manager()
                .list_table_columns(self.argument("table"))
            )

        docstring = '"""Model Definition (generated with love by Masonite) \n\n'
        for name, column in conn.items():
            length = "({})".format(column._length) if column._length else ""
            docstring += "{}: {}{} default: {}\n".format(
                name, column.get_type(), length, column.get_default()
            )

        print(docstring + '"""')



"""Base mail driver module."""

import copy
import re

from masonite.helpers import deprecated

from ...app import App
from ...helpers import config
from ...response import Responsable
from .. import BaseDriver


MAIL_FROM_RE = re.compile(r'(?:"?([^"]*)"?\s)?(?:<?(.+@[^>]+)>?)')


class BaseMailDriver(BaseDriver, Responsable):
    """Base mail driver class. This class is inherited by all mail drivers."""

    def __init__(self, app: App):
        """Base mail driver constructor.

        Arguments:
            app {masonite.app.App} -- The Masonite container class.
            view {object} -- This is the masonite.view.View class.
        """
        self.config = config("mail")
        self.app = app
        self.to_addresses = []
        self.message_subject = "Subject"
        self.message_reply_to = None
        self.from_name = self.config.FROM["name"]
        self.from_address = self.config.FROM["address"]
        self._queue = False
        self.html_content = None
        self.text_content = None
        self._message = None

    def _get_message_for_send_deprecated(self, message_contents):
        """Helper method for backwards compatibility to generate a message from .send()

        Args:
            message_contents: String

        Returns:
            message
        """
        # we used to not override self.message_body, so save it and set it back...
        old_text, old_html = self.text_content, self.html_content
        self.text_content, self.html_content = None, message_contents
        data = self.message()
        self.text_content, self.html_content = old_text, old_html
        return data

    def message(self):
        """Creates a message object for the underlying driver.

        Returns:
            message
        """
        raise NotImplementedError

    @property
    def mail_from_header(self):
        return '"{0}" <{1}>'.format(self.from_name, self.from_address)

    @property
    def mail_to_header(self):
        return ",".join(self.to_addresses)

    def text(self, content):
        """Set the text content of the email.

        Arguments:
            content {string} -- The email text content.

        Returns:
            self
        """
        self.text_content = content
        return self

    def html(self, content):
        """Set the html content of the email.

        Arguments:
            content {string} -- The email html content.

        Returns:
            self
        """
        self.html_content = content
        return self

    @property
    def message_body(self):
        """Returns the body of the message.
        """
        return self.html_content or self.text_content

    @message_body.setter
    @deprecated("Please use `.text()` and `.html()` methods instead.")
    def message_body(self, value):
        self.html_content = value

    def to(self, user_email):
        """Set the user email address who you want to send mail to.

        Arguments:
            user_email {string} -- The user email address.

        Returns:
            self
        """
        if callable(user_email):
            user_email = user_email.email

        if isinstance(user_email, (list, tuple)):
            self.to_addresses = user_email
        else:
            self.to_addresses = [user_email]
        return self

    def queue(self):
        """Whether the email should be queued or not when sending.

        Returns:
            self
        """
        self._queue = True
        return self

    def template(self, template_name, dictionary={}, mimetype="html"):
        """Create an email from a normal Jinja template.

        Arguments:
            template_name {string} -- The name of the template.

        Keyword Arguments:
            dictionary {dict} -- The data to be passed to the template. (default: {{}})
            mimetype {string} -- whether it is html or text content. (default: {html})

        Returns:
            self
        """
        view = copy.copy(self.app.make("ViewClass"))
        content = view.render(template_name, dictionary).rendered_template
        if mimetype == "html":
            self.html(content)
        else:
            self.text(content)
        return self

    def send_from(self, address, name=None):
        """Set the from address of who the sender should be.

        Arguments:
            address {string} -- A email address used as the From field in an email.
                                "John S" <john@example.com>
                                John S <john@example.com>
                                john@example.com
            name {string} -- A name used as the From field in an email.

        Returns:
            self
        """
        match = MAIL_FROM_RE.match(address)
        if not match:
            raise ValueError("Invalid address specified")

        match_name, match_address = match.groups()
        self.from_address = match_address
        if name is None and match_name:
            self.from_name = match_name

        return self

    def mailable(self, mailable):
        """Set the from address of who the sender should be.

        Arguments:
            address {string} -- A name used as the From field in an email.

        Returns:
            self
        """
        mailable = self.app.resolve(mailable.build)
        (
            self.to(mailable._to)
            .send_from(mailable._from)
            .subject(mailable._subject)
            .template(mailable.template, mailable.variables)
            .reply_to(mailable._reply_to)
        )
        return self

    def subject(self, subject):
        """Set the subject of an email.

        Arguments:
            subject {string} -- The subject of the email

        Returns:
            self
        """
        self.message_subject = subject
        return self

    def get_response(self):
        return self.message_body

    def reply_to(self, reply_to):
        """Set the Reply-To of an email.

        Arguments:
            reply_to {string} -- The reply-to of the email

        Returns:
            self
        """
        self.message_reply_to = reply_to
        return self



"""New Providers Command."""
from ..commands import BaseScaffoldCommand


class ProviderCommand(BaseScaffoldCommand):
    """
    Creates a new Service Provider.

    provider
        {name : Name of the Service Provider you want to create}
    """

    scaffold_name = "Service Provider"
    base_directory = "app/providers/"
    template = "/masonite/snippets/scaffold/provider"
    postfix = "Provider"



"""SecureHeaders Middleware."""

from ..request import Request


class SecureHeadersMiddleware:
    """SecureHeaders Middleware."""

    def __init__(self, request: Request):
        """Inject Any Dependencies From The Service Container.

        Arguments:
            Request {masonite.request.Request} -- The Masonite request object
        """
        self.request = request
        self.headers = {
            "Strict-Transport-Security": "max-age=63072000; includeSubdomains",
            "X-Frame-Options": "SAMEORIGIN",
            "X-XSS-Protection": "1; mode=block",
            "X-Content-Type-Options": "nosniff",
            "Referrer-Policy": "no-referrer, strict-origin-when-cross-origin",
            "Cache-control": "no-cache, no-store, must-revalidate",
            "Pragma": "no-cache",
        }

    def before(self):
        """Run This Middleware Before The Route Executes."""
        pass

    def after(self):
        """Run This Middleware After The Route Executes."""
        from config import middleware

        try:
            # Try importing secure headers if they exist in the config file
            self.headers.update(middleware.SECURE_HEADERS)
        except AttributeError:
            pass

        self.request.header(self.headers)



"""The RegisterController Module."""

from masonite.auth import Auth, MustVerifyEmail
from masonite.managers import MailManager
from masonite.request import Request
from masonite.validation import Validator
from masonite.view import View


class RegisterController:
    """The RegisterController class."""

    def __init__(self):
        """The RegisterController Constructor."""
        pass

    def show(self, view: View):
        """Show the registration page.

        Arguments:
            Request {masonite.request.request} -- The Masonite request class.

        Returns:
            masonite.view.View -- The Masonite View class.
        """
        return view.render("auth/register")

    def store(
        self,
        request: Request,
        mail_manager: MailManager,
        auth: Auth,
        validate: Validator,
    ):
        """Register the user with the database.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.

        Returns:
            masonite.request.Request -- The Masonite request class.
        """
        errors = request.validate(
            validate.required(["name", "email", "password"]),
            validate.email("email"),
            validate.strong(
                "password",
                length=8,
                special=1,
                uppercase=1,
                # breach=True checks if the password has been breached before.
                # Requires 'pip install pwnedapi'
                breach=False,
            ),
        )

        if errors:
            return request.back().with_errors(errors).with_input()

        user = auth.register(
            {
                "name": request.input("name"),
                "password": request.input("password"),
                "email": request.input("email"),
            }
        )

        if isinstance(user, MustVerifyEmail):
            user.verify_email(mail_manager, request)

        # Login the user
        if auth.login(request.input("email"), request.input("password")):
            # Redirect to the homepage
            return request.redirect("/home")

        # Login failed. Redirect to the register page.
        return request.back().with_input()



"""Module for Storage class."""

import os


class Storage:
    """Responsible for compiling Sass."""

    def __init__(self):
        """Storage constructor."""
        pass

    def compile_sass(self):
        """Compile sass."""
        from config import application, storage

        try:
            import sass
        except ImportError:
            pass
        else:
            matches = []
            for files in storage.SASSFILES["importFrom"]:
                for root, _, filenames in os.walk(
                    os.path.join(application.BASE_DIRECTORY, files)
                ):
                    for filename in filenames:
                        if filename.endswith(
                            (".sass", ".scss")
                        ) and not filename.startswith("_"):
                            matches.append(os.path.join(root, filename))

            for filename in matches:
                with open(filename) as f:
                    compiled_sass = sass.compile(
                        string=f.read(), include_paths=storage.SASSFILES["includePaths"]
                    )
                    name = (
                        filename.split(os.sep)[-1]
                        .replace(".scss", "")
                        .replace(".sass", "")
                    )
                    write_file = os.path.join(
                        os.path.join(
                            application.BASE_DIRECTORY, storage.SASSFILES["compileTo"]
                        ),
                        "{0}.css".format(name),
                    )
                with open(write_file, "w") as r:
                    r.write(compiled_sass)



"""List Routes Command."""
from cleo import Command
from tabulate import tabulate


class RoutesCommand(Command):
    """
    List out all routes of the application.

    show:routes
    """

    def handle(self):
        from wsgi import container

        web_routes = container.make("WebRoutes")

        routes = [["Method", "Path", "Name", "Domain", "Middleware"]]

        for route in web_routes:
            routes.append(
                [
                    route.method_type,
                    route.route_url,
                    route.named_route,
                    route.required_domain,
                    ",".join(route.list_middleware),
                ]
            )

        print(tabulate(routes, headers="firstrow", tablefmt="rst"))



"""Time Module."""

import pendulum


def cookie_expire_time(str_time):
    """Take a string like 1 month or 5 minutes and returns a pendulum instance.

    Arguments:
        str_time {string} -- Could be values like 1 second or 3 minutes

    Returns:
        pendulum -- Returns Pendulum instance
    """
    if str_time != "expired":
        number = int(str_time.split(" ")[0])
        length = str_time.split(" ")[1]

        if length in ("second", "seconds"):
            return (
                pendulum.now("GMT")
                .add(seconds=number)
                .format("%a, %d %b %Y %H:%M:%S GMT")
            )
        elif length in ("minute", "minutes"):
            return (
                pendulum.now("GMT")
                .add(minutes=number)
                .format("%a, %d %b %Y %H:%M:%S GMT")
            )
        elif length in ("hour", "hours"):
            return (
                pendulum.now("GMT")
                .add(hours=number)
                .format("%a, %d %b %Y %H:%M:%S GMT")
            )
        elif length in ("days", "days"):
            return (
                pendulum.now("GMT").add(days=number).format("%a, %d %b %Y %H:%M:%S GMT")
            )
        elif length in ("week", "weeks"):
            return pendulum.now("GMT").add(weeks=1).format("%a, %d %b %Y %H:%M:%S GMT")
        elif length in ("month", "months"):
            return (
                pendulum.now("GMT")
                .add(months=number)
                .format("%a, %d %b %Y %H:%M:%S GMT")
            )
        elif length in ("year", "years"):
            return (
                pendulum.now("GMT")
                .add(years=number)
                .format("%a, %d %b %Y %H:%M:%S GMT")
            )

        return None
    else:
        return (
            pendulum.now("GMT").subtract(years=20).format("%a, %d %b %Y %H:%M:%S GMT")
        )


def parse_human_time(str_time):
    """Take a string like 1 month or 5 minutes and returns a pendulum instance.

    Arguments:
        str_time {string} -- Could be values like 1 second or 3 minutes

    Returns:
        pendulum -- Returns Pendulum instance
    """
    if str_time != "expired":
        number = int(str_time.split(" ")[0])
        length = str_time.split(" ")[1]

        if length in ("second", "seconds"):
            return pendulum.now("GMT").add(seconds=number)
        elif length in ("minute", "minutes"):
            return pendulum.now("GMT").add(minutes=number)
        elif length in ("hour", "hours"):
            return pendulum.now("GMT").add(hours=number)
        elif length in ("days", "days"):
            return pendulum.now("GMT").add(days=number)
        elif length in ("week", "weeks"):
            return pendulum.now("GMT").add(weeks=1)
        elif length in ("month", "months"):
            return pendulum.now("GMT").add(months=number)
        elif length in ("year", "years"):
            return pendulum.now("GMT").add(years=number)

        return None
    else:
        return pendulum.now("GMT").subtract(years=20)



"""Third party package integrations."""
import os
import shutil
import sys


def create_or_append_config(location, name=False):
    if name:
        file_name = name
    else:
        file_name = os.path.basename(location)

    # import it into the config directory
    config_directory = os.path.join(os.getcwd(), "config")

    # if file does not exist
    if not os.path.isfile(config_directory + "/" + file_name):
        shutil.copyfile(location, config_directory + "/" + file_name)
        print("\033[92mConfiguration File Created!\033[0m")
    else:
        # Append to the file
        with open(config_directory + "/" + file_name, "a") as project_config, open(
            location, "r"
        ) as package_config:
            project_config.write(package_config.read())

        print("\033[92mConfiguration File Appended!\033[0m")


def append_web_routes(location):
    # import it into the web.py file
    routes_file = os.path.join(os.getcwd(), "routes/web.py")

    with open(routes_file, "a") as project_routes, open(
        location, "r"
    ) as package_routes:
        project_routes.write(package_routes.read())

    print("\033[92mroutes/web.py File Appended!\033[0m")


def append_file(from_location, to_location):
    with open(from_location, "r") as from_file_pointer, open(
        os.path.join(os.getcwd(), to_location), "a"
    ) as to_file_pointer:
        to_file_pointer.write(from_file_pointer.read())

    print("\033[92m {} has been appended! \033[0m".format(to_location))


def append_api_routes(location):
    # import it into the web.py file
    api_file = os.path.join(os.getcwd(), "routes/api.py")

    # Append to the file
    with open(api_file, "a") as project_routes, open(location, "r") as package_routes:
        project_routes.write(package_routes.read())

    print("\033[92mroutes/api.py File Appended!\033[0m")


def create_controller(location, to="app/http/controllers"):
    file_name = os.path.basename(location)

    controller_directory = os.path.join(os.getcwd(), to)
    controller_file = os.path.join(controller_directory, file_name)
    if not os.path.exists(controller_directory):
        # Create the path to the model if it does not exist
        os.makedirs(controller_directory)

    if os.path.isfile(controller_file):
        # if file does exist
        print("\033[91m{0} Controller Already Exists!\033[0m".format(file_name))
    else:
        # copy controller over
        shutil.copyfile(location, controller_file)

        print("\033[92m{0} Controller Created\033[0m".format(file_name))


def add_venv_site_packages():
    try:
        from config import packages

        # Add additional site packages to vendor if they exist
        for directory in packages.SITE_PACKAGES:
            path = os.path.join(os.getcwd(), directory)
            sys.path.append(path)
    except ImportError:
        raise ImportError

    if "VIRTUAL_ENV" in os.environ:
        python_version = None
        venv_directory = os.listdir(os.path.join(os.environ["VIRTUAL_ENV"], "lib"))

        for directory in venv_directory:
            if directory.startswith("python"):
                python_version = directory
                break

        if python_version:
            site_packages_directory = os.path.join(
                os.environ["VIRTUAL_ENV"], "lib", python_version, "site-packages"
            )

            sys.path.append(site_packages_directory)
        else:
            print(
                "\033[93mWARNING: Could not add the virtual environment you are currently in. Attempting to add: {0}\033[93m".format(
                    os.environ["VIRTUAL_ENV"]
                )
            )


class PackageContainer:
    def create(self):
        from masonite.app import App
        from config import providers

        container = App()

        container.bind("Container", container)

        container.bind("ProvidersConfig", providers)
        container.bind("Providers", [])
        container.bind("WSGIProviders", [])

        for provider in container.make("ProvidersConfig").PROVIDERS:
            located_provider = provider()
            located_provider.load_app(container).register()
            if located_provider.wsgi:
                container.make("WSGIProviders").append(located_provider)
            else:
                container.make("Providers").append(located_provider)

        for provider in container.make("Providers"):
            container.resolve(provider.boot)

        return container



import os
import sys
from pydoc import ErrorDuringImport
from cleo import Application
from .commands import NewCommand, InstallCommand
from . import __version__

sys.path.append(os.getcwd())

application = Application("Masonite Version:", __version__)
application.add(NewCommand())
application.add(InstallCommand())


try:
    from wsgi import container
    from cleo import Command

    for key, value in container.collect(Command).items():
        application.add(value)
except ErrorDuringImport as e:
    print(e)
except ImportError:
    pass

if __name__ == "__main__":
    application.run()



import os
import shutil
import json


class Preset:
    def ensure_component_directory_exists(self):
        """Ensure the component directories we need exist."""
        directory = "resources/js/components"
        if not os.path.exists(os.path.realpath(directory)):
            os.makedirs(os.path.realpath(directory))

    def update_packages(self, dev=True):
        """Update the "package.json" file."""
        if not os.path.exists(os.path.realpath("package.json")):
            return

        configuration_key = "devDependencies" if dev else "dependencies"

        packages = {}
        with open(os.path.realpath("package.json"), "r+") as f:
            packages = json.load(f)
            packages[configuration_key] = self.update_package_array(
                packages[configuration_key] if configuration_key in packages else {}
            )
            f.seek(0)  # Rewind to beginning of file
            f.truncate()
            f.write(json.dumps(packages, sort_keys=True, indent=4))

    def remove_node_modules(self):
        """Remove the installed Node modules."""
        for filename in ["package-lock.json", "yarn.lock"]:
            if os.path.exists(os.path.realpath(filename)):
                os.remove(filename)
        shutil.rmtree("node_modules", ignore_errors=True)

    def create_scss_file(self):
        """Create an empty app.scss file"""
        os.makedirs(os.path.realpath("resources/sass"))
        with open(os.path.realpath("resources/sass/app.scss"), "w") as f:
            f.write("// Add your Sass here")



"""Run Seed Command."""
import subprocess

from cleo import Command


class SeedRunCommand(Command):
    """
    Run seed for database.

    seed:run
        {table=None : Name of the table to seed}
    """

    def handle(self):

        table = self.argument("table").lower()
        if not table == "none":
            seeder = "--seeder {}_table_seeder".format(table.lower())
        else:
            seeder = ""

        subprocess.call(
            [
                "orator db:seed -p databases/seeds -c config/database.py -f {}".format(
                    seeder
                ),
            ],
            shell=True,
        )



"""AuthJWTDriver Module."""

import pendulum
from ...auth import Auth
from ...contracts import AuthContract
from ...drivers import BaseDriver
from ...exceptions import DriverLibraryNotFound
from ...helpers import config, cookie_expire_time
from ...request import Request


class AuthJwtDriver(BaseDriver, AuthContract):
    def __init__(self, request: Request):
        """AuthCookieDriver initializer.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
        """
        self.request = request
        try:
            import jwt

            self.jwt = jwt
        except ImportError:
            raise DriverLibraryNotFound(
                "Please install pyjwt by running 'pip install pyjwt'"
            )

    def user(self, auth_model):
        """Gets the user based on this driver implementation

        Arguments:
            auth_model {orator.orm.Model} -- An Orator ORM type object.

        Returns:
            Model|bool
        """
        from config.application import KEY

        if self.request.get_cookie("token"):

            try:
                token = self.jwt.decode(
                    self.request.get_cookie("token"), KEY, algorithms=["HS256"]
                )
            except self.jwt.exceptions.DecodeError:
                self.delete()
                return False

            expired = token["expired"]
            token.pop("expired")
            if not pendulum.parse(expired).is_past():
                auth_model = auth_model()
                auth_model.fill(**token)
                return auth_model

            if config("auth.drivers.jwt.reauthentication", True):
                auth_model = Auth(self.request).login_by_id(
                    token[auth_model.__primary_key__]
                )
            else:
                auth_model.fill(**token)

            token.update(
                {
                    "expired": cookie_expire_time(
                        config("auth.drivers.jwt.lifetime", "5 minutes")
                    )
                }
            )
            self.request.cookie("token", token)
            return auth_model
        return False

    def save(self, _, **kwargs):
        """Saves the state of authentication.

        In this case the state is serializing the user model and saving to a token cookie.

        Arguments:
            remember_token {string} -- A token containing the state.

        Returns:
            bool
        """
        from config.application import KEY

        model = kwargs.get("model", False)
        serialized_dictionary = model.serialize()
        serialized_dictionary.update({"expired": cookie_expire_time("5 minutes")})
        token = self.jwt.encode(serialized_dictionary, KEY, algorithm="HS256")
        if isinstance(token, bytes):
            token = bytes(token).decode("utf-8")
        self.request.cookie("token", token)

    def delete(self):
        """Deletes the state depending on the implementation of this driver.

        Returns:
            bool
        """
        self.request.delete_cookie("token")

    def logout(self):
        self.delete()
        self.request.reset_user()



"""New Model Command."""
from ..app import App
from ..helpers.filesystem import make_directory
from ..view import View

from cleo import Command


class ModelCommand(Command):
    """
    Creates a model.

    model
        {name : Name of the model}
        {--m|migration : Create a migration for specified model}
        {--s|seed=? : Create a database seed}
    """

    scaffold_name = "Model"
    template = "/masonite/snippets/scaffold/model"
    base_directory = "app/"

    def handle(self):
        class_name = self.argument("name")
        view = View(App())
        class_directory = "{}{}.py".format(self.base_directory, class_name)

        if not make_directory(class_directory):
            return self.error("{0} Already Exists!".format(self.scaffold_name))

        with open(class_directory, "w+") as f:
            if view.exists(self.template):
                f.write(
                    view.render(
                        self.template, {"class": class_name.split("/")[-1]}
                    ).rendered_template
                )
                self.info("{} Created Successfully!".format(self.scaffold_name))

        if self.option("migration"):
            model_name = class_name.lower() + "s"
            self.call(
                "migration",
                [("name", "create_{}_table".format(model_name)), ("-c", model_name)],
            )

        if self.option("seed"):
            seed_file = model_name
            seed_file = self.option("seed")

            self.call("seed", [("table", seed_file)])



"""Module for the Pusher websocket driver."""

from ...contracts import BroadcastContract
from ...drivers import BaseDriver
from ...exceptions import DriverLibraryNotFound
from ...helpers import config


class BroadcastPusherDriver(BroadcastContract, BaseDriver):
    """Class for the Pusher websocket driver."""

    def __init__(self):
        """Pusher driver constructor.

        Arguments:
            BroadcastConfig {config.broadcast} -- Broadcast configuration.
        """
        self.ssl_message = True

    def ssl(self, boolean):
        """Set whether to send data with SSL enabled.

        Arguments:
            boolean {bool} -- Boolean on whether to set SSL.

        Returns:
            self
        """
        self.ssl_message = boolean
        return self

    def channel(self, channels, message, event="base-event"):
        """Specify which channel(s) you want to send information to.

        Arguments:
            channels {string|list} -- Can be a string for the channel or a list of strings for the channels.
            message {string} -- The message you want to send to the channel(s)

        Keyword Arguments:
            event {string} -- The event you want broadcasted along with your data. (default: {'base-event'})

        Raises:
            DriverLibraryNotFound -- Thrown when pusher is not installed.

        Returns:
            string -- Returns the message sent.
        """
        try:
            import pusher
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "pusher" library. Please pip install this library running "pip install pusher"'
            )

        configuration = config("broadcast.drivers.pusher")

        pusher_client = pusher.Pusher(
            app_id=str(configuration["app_id"]),
            key=configuration["client"],
            secret=configuration["secret"],
            ssl=self.ssl_message,
        )

        if isinstance(message, str):
            message = {"message": message}

        if isinstance(channels, list):
            for channel in channels:
                pusher_client.trigger(channel, event, message)
        else:
            pusher_client.trigger(channels, event, message)

        return message



class Mailable:

    _to = ""
    _from = ""
    _subject = ""

    def view(self, template, variables={}):
        self.template = template
        self.variables = variables
        return self

    def to(self, to):
        self._to = to
        return self

    def reply_to(self, reply_to):
        self._reply_to = reply_to
        return self

    def send_from(self, send_from):
        self._from = send_from
        return self

    def subject(self, subject):
        self._subject = subject
        return self



from cleo import Command

from ..app import App
from ..helpers.filesystem import make_directory
from ..view import View


class BaseScaffoldCommand(Command):
    """
    Creates a model.

    model
        {name : Name of the model}
    """

    scaffold_name = "Example"
    suffix = ""
    postfix = ""
    prefix = ""
    file_extension = ".py"
    base_directory = "app/example/"
    file_to_lower = False

    template = "/masonite/snippets/scaffold/model"

    def handle(self):
        # If postfix already exists as part of the name, don't add it again
        if self.postfix and self.argument("name").lower().endswith(
            self.postfix.lower()
        ):
            class_name = self.argument("name")
        else:
            class_name = self.argument("name") + self.postfix

        view = View(App())
        class_directory = "{}{}{}{}".format(
            self.base_directory, class_name, self.suffix, self.file_extension
        )

        if self.file_to_lower:
            class_directory = class_directory.lower()

        if not make_directory(class_directory):
            return self.error("{0} Already Exists!".format(self.scaffold_name))

        with open(class_directory, "w+") as f:
            if view.exists(self.template):
                f.write(
                    view.render(
                        self.template,
                        {"class": self.prefix + class_name.split("/")[-1]},
                    ).rendered_template
                )
                self.info("{} Created Successfully!".format(self.scaffold_name))



"""New Seeder Command."""
import subprocess
import os

from cleo import Command


class SeedCommand(Command):
    """
    Create a seeder to seed a database.

    seed
        {table : Name of the table to seed}
    """

    def handle(self):
        table = self.argument("table").lower()
        subprocess.call(
            ["orator make:seed {}_table_seeder -p databases/seeds".format(table)],
            shell=True,
        )

        self.check_init_file()

    def check_init_file(self):
        os.makedirs(os.path.dirname("databases/seeds/__init__.py"), exist_ok=True)

        with open("databases/seeds/__init__.py") as f:
            if "sys.path.append(os.getcwd())" not in f.read():
                with open("databases/seeds/__init__.py", "w+") as fp:
                    fp.write("import os\nimport sys\nsys.path.append(os.getcwd())\n")



"""Log Driver Module."""

import logging
import os

from ...app import App
from ...contracts import MailContract
from ...drivers import BaseMailDriver


class MailLogDriver(BaseMailDriver, MailContract):
    """Mail log driver."""

    def __init__(self, app: App):
        super().__init__(app)

        if "log" in self.config.DRIVERS and "location" in self.config.DRIVERS["log"]:
            log_location = self.config.DRIVERS["log"]["location"]
        else:
            log_location = "bootstrap/mail"

        if not os.path.exists(log_location):
            # Create the path to the model if it does not exist
            os.makedirs(log_location)

        handler = logging.FileHandler(
            "{0}/{1}".format(log_location, os.getenv("MAIL_LOGFILE", "mail.log")),
            delay=True,
        )
        self.logger = logging.getLogger(__name__)
        self.logger.handlers = []
        self.logger.propagate = False
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)

    def send(self, message=None):
        """Prints the message in a log.

        Keyword Arguments:
            message {string} -- The message to be printed. (default: { None })

        Returns:
            None
        """

        if not message:
            message = self.message_body

        self.logger.info("***************************************")

        self.logger.info("To: {}".format(self.mail_to_header))
        self.logger.info("From: {}".format(self.mail_from_header))
        self.logger.info("Subject: {}".format(self.message_subject))
        self.logger.info("Reply-To: {}".format(self.message_reply_to))
        self.logger.info("Message: ")
        self.logger.info(message)

        self.logger.info("***************************************")



"""New Job Command."""
from ..commands import BaseScaffoldCommand


class MailableCommand(BaseScaffoldCommand):
    """
    Creates a new Mailable.

    mailable
        {name : Name of the job you want to create}
    """

    scaffold_name = "Mailable"
    template = "/masonite/snippets/scaffold/mailable"
    base_directory = "app/mailable/"
    postfix = "Mailable"



"""A PasswordController Module."""

import uuid

from masonite import env, Mail, Session
from masonite.auth import Auth
from masonite.helpers import config, password as bcrypt_password
from masonite.request import Request
from masonite.view import View
from masonite.validation import Validator
from config.auth import AUTH


class PasswordController:
    """Password Controller."""

    def forget(self, view: View, auth: Auth):
        return view.render("auth/forget", {"app": config("application"), "Auth": auth})

    def reset(self, view: View, request: Request, auth: Auth):
        token = request.param("token")
        user = AUTH["guards"]["web"]["model"].where("remember_token", token).first()
        if user:
            return view.render(
                "auth/reset",
                {"token": token, "app": config("application"), "Auth": auth},
            )

    def send(self, request: Request, session: Session, mail: Mail, validate: Validator):
        errors = request.validate(validate.required("email"), validate.email("email"))

        if errors:
            return request.back().with_errors(errors)

        email = request.input("email")
        user = AUTH["guards"]["web"]["model"].where("email", email).first()

        if user:
            if not user.remember_token:
                user.remember_token = str(uuid.uuid4())
                user.save()
            message = "Please visit {}/password/{}/reset to reset your password".format(
                env("SITE", "http://localhost:8000"), user.remember_token
            )
            mail.subject("Reset Password Instructions").to(user.email).send(message)

        session.flash(
            "success",
            "If we found that email in our system then the email has been sent. Please follow the instructions in the email to reset your password.",
        )
        return request.redirect("/password")

    def update(self, request: Request, validate: Validator):
        errors = request.validate(
            validate.required("password"),
            # TODO: only available in masonite latest versions (which are not compatible with Masonite 2.2)
            validate.strong(
                "password",
                length=8,
                special=1,
                uppercase=1,
                # breach=True checks if the password has been breached before.
                # Requires 'pip install pwnedapi'
                breach=False,
            ),
        )

        if errors:
            return request.back().with_errors(errors)

        user = (
            AUTH["guards"]["web"]["model"]
            .where("remember_token", request.param("token"))
            .first()
        )
        if user:
            user.password = bcrypt_password(request.input("password"))
            user.save()
            return request.redirect("/login")



from orator.migrations import Migration


class CreateQueueJobsTable(Migration):
    def up(self):
        """Run the migrations."""
        with self.schema.create("queue_jobs") as table:
            table.increments("id")
            table.string("name")
            table.binary("serialized")
            table.integer("attempts")
            table.integer("failed").nullable()
            table.timestamp("ran_at").nullable()
            table.timestamp("created_at").nullable()
            table.timestamp("wait_until").nullable()

    def down(self):
        """Revert the migrations."""
        self.schema.drop("queue_jobs")



from .static import static
from .password import password
from .misc import (
    random_string,
    dot,
    clean_request_input,
    HasColoredCommands,
    Compact as compact,
    deprecated,
)
from .Extendable import Extendable
from .time import cookie_expire_time, parse_human_time
from .optional import Optional as optional
from .structures import config, Dot, load
from .migrations import has_unmigrated_migrations
from orator.support.collection import Collection as collect



"""A StatusProvider Service Provider."""

import json

from ..helpers import config
from ..provider import ServiceProvider
from ..response import Response


class ServerErrorExceptionHook:
    def load(self, app):
        from config import application

        if application.DEBUG:
            return

        request = app.make("Request")

        request.status(500)
        if app.make("ViewClass").exists("errors/500"):
            rendered_view = app.make("View")("errors/500")
        else:
            rendered_view = app.make("View")(
                config(
                    "application.templates.statuscode", "/masonite/snippets/statuscode"
                ),
                {"code": "500 Internal Server Error"},
            )

        request.app().make(Response).view(rendered_view)


class StatusCodeProvider(ServiceProvider):
    def register(self):
        self.app.bind("ServiceErrorExceptionHook", ServerErrorExceptionHook())

    def boot(self):
        request = self.app.make("Request")
        response = self.app.make(Response)
        if request.is_status(200):
            return

        if request.get_status() in (500, 405, 404):
            if "application/json" in request.header("Content-Type"):
                # Returns json response when we want the client to receive a json response
                body = json.loads(self.app.make("Response").decode("utf-8"))
                json_response = {
                    "error": {"status": request.get_status(), "body": body}
                }
                response.view(json_response, status=request.get_status())
            else:
                # Returns html response when json is not explicitly specified
                if self.app.make("ViewClass").exists(
                    "errors/{}".format(request.get_status())
                ):
                    rendered_view = self.app.make("View")(
                        "errors/{}".format(request.get_status())
                    )
                else:
                    rendered_view = self.app.make("View")(
                        config(
                            "application.templates.statuscode",
                            "/masonite/snippets/statuscode",
                        ),
                        {"code": request.get_status_code()},
                    )

                response.view(rendered_view, status=request.get_status())



from cleo import Command
import os
import shutil
import zipfile
from ..exceptions import ProjectLimitReached


class NewCommand(Command):
    """
    Creates a new Masonite project

    new
        {name? : Name of your Masonite project}
        {target? : Path of you Masonite project}
        {--b|--branch=False : Specify which branch from the Masonite repo you would like to install}
        {--r|--release=False : Specify which version of Masonite you would like to install}
    """

    def handle(self):
        name = self.argument("name")
        if not name:
            name = "project"

        target = self.argument("target")

        if not target:
            target = "."

        branch = self.option("branch")
        version = self.option("release")
        if not os.path.isdir(os.path.join(os.getcwd(), name)):
            from io import BytesIO
            import requests

            for directory in os.listdir(os.getcwd()):
                if directory.startswith("masonite-"):
                    return self.comment(
                        'There is a folder that starts with "masonite-" and therefore craft cannot create a new project'
                    )
            try:
                if branch != "False":
                    get_branch = requests.get(
                        "https://api.github.com/repos/MasoniteFramework/cookie-cutter/branches/{0}".format(
                            branch
                        )
                    )

                    if "name" not in get_branch.json():
                        return self.comment("Branch {0} does not exist.".format(branch))

                    zipball = "https://github.com/MasoniteFramework/cookie-cutter/archive/{0}.zip".format(
                        branch
                    )
                elif version != "False":
                    get_zip_url = requests.get(
                        "https://api.github.com/repos/MasoniteFramework/cookie-cutter/releases"
                    )
                    zipball = False

                    for release in get_zip_url.json():
                        if "tag_name" in release and release["tag_name"].startswith(
                            "v{0}".format(version)
                        ):
                            self.info(
                                "Installing version {0}".format(release["tag_name"])
                            )
                            self.line("")
                            zipball = release["zipball_url"]
                            break

                    if zipball is False:
                        return self.info(
                            "Version {0} could not be found".format(version)
                        )
                else:
                    get_zip_url = requests.get(
                        "https://api.github.com/repos/MasoniteFramework/cookie-cutter/releases"
                    )
                    tags = []

                    for release in get_zip_url.json():
                        if release["prerelease"] is False:
                            tags.append(release["tag_name"].replace("v", ""))

                    tags = sorted(
                        tags, key=lambda v: [int(i) for i in v.split(".")], reverse=True
                    )

                    get_zip_url = requests.get(
                        "https://api.github.com/repos/MasoniteFramework/cookie-cutter/releases/tags/v{0}".format(
                            tags[0]
                        )
                    )

                    zipball = get_zip_url.json()["zipball_url"]
            except TypeError:
                raise ProjectLimitReached(
                    "You have reached your hourly limit of creating new projects. Try again in 1 hour."
                )
            success = False

            zipurl = zipball

            self.info("Crafting Application ...")

            try:
                # Python 3
                from urllib.request import urlopen

                with urlopen(zipurl) as zipresp:
                    with zipfile.ZipFile(BytesIO(zipresp.read())) as zfile:
                        zfile.extractall(os.getcwd())

                success = True
            except ImportError:
                # Python 2
                import urllib

                r = urllib.urlopen(zipurl)
                with zipfile.ZipFile(BytesIO(r.read())) as z:
                    z.extractall(os.getcwd())

                success = True
            except Exception as e:
                raise e

            if success:
                for directory in os.listdir(os.getcwd()):
                    if directory.startswith(
                        "MasoniteFramework-cookie-cutter"
                    ) or directory.startswith("cookie-cutter-"):
                        if target == ".":
                            from_dir = os.path.join(
                                os.getcwd(), "{0}".format(directory)
                            )
                            to_dir = os.path.abspath(os.path.expanduser(target))

                            for file in os.listdir(from_dir):
                                shutil.move(os.path.join(from_dir, file), to_dir)
                            os.rmdir(from_dir)
                        else:
                            os.rename(
                                os.path.join(os.getcwd(), "{0}".format(directory)),
                                os.path.join(os.getcwd(), target),
                            )

                        self.info("Application Created Successfully!")
                        self.info("Installing Dependencies ")
                        if target == ".":
                            self.call("install")

                            self.info(
                                "Installed Successfully. Just Run `craft serve` To Start Your Application."
                            )
                        else:
                            self.info(
                                "Project Created Successfully. You now will have to CD into your new '{}' directory and run `craft install` to complete the installation".format(
                                    target
                                )
                            )

                        return

            else:
                self.comment("Could Not Create Application :(")
        else:
            self.comment(
                "Directory {0} already exists. Please choose another project name".format(
                    name
                )
            )



import io


class MockWsgiInput:
    def __init__(self, data):
        self.data = data

    def read(self, _):
        return self.data


def generate_wsgi():
    return {
        "wsgi.version": (1, 0),
        "wsgi.multithread": False,
        "wsgi.multiprocess": True,
        "wsgi.run_once": False,
        "wsgi.input": io.BytesIO(),
        "SERVER_SOFTWARE": "gunicorn/19.7.1",
        "REQUEST_METHOD": "GET",
        "QUERY_STRING": "application=Masonite",
        "RAW_URI": "/",
        "SERVER_PROTOCOL": "HTTP/1.1",
        "HTTP_HOST": "127.0.0.1:8000",
        "HTTP_ACCEPT": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        "HTTP_UPGRADE_INSECURE_REQUESTS": "1",
        "HTTP_COOKIE": "setcookie=value",
        "HTTP_USER_AGENT": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_2) AppleWebKit/604.4.7 (KHTML, like Gecko) Version/11.0.2 Safari/604.4.7",
        "HTTP_ACCEPT_LANGUAGE": "en-us",
        "HTTP_ACCEPT_ENCODING": "gzip, deflate",
        "HTTP_CONNECTION": "keep-alive",
        "wsgi.url_scheme": "http",
        "REMOTE_ADDR": "127.0.0.1",
        "REMOTE_PORT": "62241",
        "SERVER_NAME": "127.0.0.1",
        "SERVER_PORT": "8000",
        "PATH_INFO": "/",
        "SCRIPT_NAME": "",
    }



"""Migrate Status Command."""
import os
import sys

from subprocess import check_output
from cleo import Command
from ..packages import add_venv_site_packages


class MigrateStatusCommand(Command):
    """
    Migrate status.

    migrate:status
    """

    def handle(self):
        sys.path.append(os.getcwd())
        try:
            add_venv_site_packages()
        except ImportError:
            self.comment(
                "This command must be ran inside of the root of a Masonite project directory"
            )

        from wsgi import container

        migration_directory = ["databases/migrations"]
        for key, value in container.providers.items():
            if isinstance(key, str) and "MigrationDirectory" in key:
                migration_directory.append(value)

        for directory in migration_directory:
            self.line("")
            if len(migration_directory) > 1:
                self.info("Migrate Status: {}".format(directory))
            try:
                output = bytes(
                    check_output(
                        [
                            "orator",
                            "migrate:status",
                            "-c",
                            "config/database.py",
                            "-p",
                            directory,
                        ]
                    )
                ).decode("utf-8")

                self.line(
                    output.replace("Yes", "<info>Yes</info>").replace(
                        "No", "<comment>No</comment>"
                    )
                )
            except Exception:
                pass



"""A LoginController Module."""

from masonite.auth import Auth
from masonite.request import Request
from masonite.validation import Validator
from masonite.view import View


class LoginController:
    """Login Form Controller."""

    def __init__(self):
        """LoginController Constructor."""
        pass

    def show(self, request: Request, view: View):
        """Show the login page.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
            view {masonite.view.View} -- The Masonite view class.

        Returns:
            masonite.view.View -- Returns the Masonite view class.
        """
        if request.user():
            return request.redirect("/home")

        return view.render("auth/login")

    def store(self, request: Request, auth: Auth, validate: Validator):
        """Login the user.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
            auth {masonite.auth.auth} -- The Masonite auth class.
            validate {masonite.validator.Validator} -- The Masonite Validator class.

        Returns:
            masonite.request.Request -- The Masonite request class.
        """
        errors = request.validate(
            validate.required(["email", "password"]), validate.email("email"),
        )

        if errors:
            return request.back().with_errors(errors).with_input()

        if auth.login(request.input("email"), request.input("password")):
            return request.redirect("/home")

        return request.back().with_errors({"email": ["Email or password is incorrect"]})

    def logout(self, request: Request, auth: Auth):
        """Log out the user.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
            auth {masonite.auth.auth} -- The Masonite auth class.

        Returns:
            masonite.request.Request -- The Masonite request class.
        """
        auth.logout()
        return request.redirect("/login")



"""The HomeController Module."""

from masonite.auth import Auth
from masonite.request import Request
from masonite.view import View


class HomeController:
    """Home Dashboard Controller."""

    def __init__(self):
        pass

    def show(self, request: Request, view: View, auth: Auth):
        if not auth.user():
            request.redirect("/login")
        return view.render("auth/home")



"""Mailgun Driver Module."""
import warnings

import requests

from ...contracts.MailContract import MailContract
from ...drivers import BaseMailDriver


class MailMailgunDriver(BaseMailDriver, MailContract):
    """Mailgun driver."""

    def message(self):
        data = {
            "from": self.mail_from_header,
            "to": self.to_addresses,
            "subject": self.message_subject,
            "h:Reply-To": self.message_reply_to,
        }

        # Attach both mimetypes if they exist.
        if self.text_content:
            data["text"] = self.text_content
        if self.html_content:
            data["html"] = self.html_content

        return data

    def send(self, message=None):
        """Send the message through the Mailgun service.

        Keyword Arguments:
            message {string} -- The message to be sent to Mailgun. (default: {None})

        Returns:
            requests.post -- Returns the response as a requests object.
        """
        if message and isinstance(message, str):
            warnings.warn(
                "Passing message to .send() is deprecated. Please use .text() and .html().",
                category=DeprecationWarning,
                stacklevel=2,
            )
            data = self._get_message_for_send_deprecated(message)

        # The above should be removed once deprecation time period passed.
        elif not message:
            data = self.message()
        else:
            data = message

        if self._queue:
            from wsgi import container
            from ... import Queue

            return container.make(Queue).push(self._send_mail, args=(data,))

        return self._send_mail(data)

    def _send_mail(self, data):
        """Wrapper around sending mail so it can also be used with queues.

        Arguments:
            data {dict} -- The data for mailgun post request.

        Returns:
            requests.post
        """

        domain = self.config.DRIVERS["mailgun"]["domain"]
        secret = self.config.DRIVERS["mailgun"]["secret"]

        return requests.post(
            "https://api.mailgun.net/v3/{0}/messages".format(domain),
            auth=("api", secret),
            data=data,
        )



"""Terminal Driver Module."""

import logging

from ...app import App
from ...contracts.MailContract import MailContract
from ...drivers import BaseMailDriver


class MailTerminalDriver(BaseMailDriver, MailContract):
    """Mail terminal driver."""

    def __init__(self, app: App):
        super().__init__(app)
        self.logger = logging.getLogger(__name__)
        self.logger.handlers = []
        handler = logging.StreamHandler()
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(handler)
        self.logger.propagate = False

    def send(self, message=None):
        """Prints the message to the terminal.

        Keyword Arguments:
            message {string} -- The message to be printed. (default: { None })

        Returns:
            None
        """

        if not message:
            message = self.message_body

        self.logger.info("***************************************")

        self.logger.info("To: {}".format(self.mail_to_header))
        self.logger.info("From: {}".format(self.mail_from_header))
        self.logger.info("Subject: {}".format(self.message_subject))
        self.logger.info("Reply-To: {}".format(self.message_reply_to))
        self.logger.info("Message: ")
        self.logger.info(message)

        self.logger.info("***************************************")



"""New Migration Command."""
import subprocess

from cleo import Command


class MakeMigrationCommand(Command):
    """
    Makes a new migration.

    migration
        {name : Name of your migration}
        {--t|--table=False : Table you are migrating for}
        {--c|--create=False : Table you want to create with this migration}
    """

    def handle(self):
        name = self.argument("name")

        if self.option("create") != "False":
            subprocess.call(
                [
                    "orator",
                    "make:migration",
                    name,
                    "-p",
                    "databases/migrations",
                    "--table",
                    self.option("create"),
                    "--create",
                ]
            )
        elif self.option("table") != "False":
            subprocess.call(
                [
                    "orator",
                    "make:migration",
                    name,
                    "-p",
                    "databases/migrations",
                    "--table",
                    self.option("table"),
                ]
            )
        else:
            subprocess.call(
                ["orator", "make:migration", name, "-p", "databases/migrations"]
            )



"""A Csrf Service Provider."""

from ..auth import Csrf
from ..provider import ServiceProvider


class CsrfProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind("Csrf", Csrf(self.app.make("Request")))

    def boot(self):
        pass



"""New Controller Command."""
from ..view import View
from ..app import App
from ..helpers.filesystem import make_directory

from cleo import Command


class ControllerCommand(Command):
    """
    Creates a controller.

    controller
        {name : Name of the controller you would like to create}
        {--r|--resource : Create a controller as a resource}
        {--e|--exact : For add the name controller without `Controller` text}
    """

    def handle(self):
        controller = self.argument("name")
        view = View(App())

        if not self.option("exact"):
            controller = controller + "Controller"

        if not make_directory("app/http/controllers/{0}.py".format(controller)):
            return self.error("{0} Controller Exists!".format(controller))

        with open("app/http/controllers/{0}.py".format(controller), "w+") as f:
            if view.exists("/masonite/snippets/scaffold/controller"):
                if self.option("resource"):
                    template = "/masonite/snippets/scaffold/controller_resource"
                else:
                    template = "/masonite/snippets/scaffold/controller"

                f.write(
                    view.render(
                        template, {"class": controller.split("/")[-1]}
                    ).rendered_template
                )

        self.info("Controller Created Successfully!")



"""A WhiteNoiseProvider Service Provider."""

from whitenoise import WhiteNoise

from ..provider import ServiceProvider
from ..helpers import config


class WhitenoiseProvider(ServiceProvider):

    wsgi = False

    def register(self):
        pass

    def boot(self):
        """Wrap the WSGI server in a whitenoise container."""
        from config import application

        self.app.bind(
            "WSGI",
            WhiteNoise(
                self.app.make("WSGI"),
                root=config("application.static_root"),
                autorefresh=application.DEBUG,
            ),
        )

        for location, alias in self.app.make("staticfiles").items():
            self.app.make("WSGI").add_files(location, prefix=alias)



"""AuthCookieDriver Module."""

from ...contracts import AuthContract
from ...drivers import BaseDriver
from ...request import Request


class AuthCookieDriver(BaseDriver, AuthContract):
    def __init__(self, request: Request):
        """AuthCookieDriver initializer.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
        """
        self.request = request

    def user(self, auth_model):
        """Gets the user based on this driver implementation

        Arguments:
            auth_model {orator.orm.Model} -- An Orator ORM type object.

        Returns:
            Model|bool
        """
        if self.request.get_cookie("token") and auth_model:
            return auth_model.where(
                "remember_token", self.request.get_cookie("token")
            ).first()

        return False

    def save(self, remember_token, **_):
        """Saves the cookie to some state.

        In this case the state is saving to a cookie.

        Arguments:
            remember_token {string} -- A token containing the state.

        Returns:
            bool
        """
        return self.request.cookie("token", remember_token)

    def delete(self):
        """Deletes the state depending on the implementation of this driver.

        Returns:
            bool
        """
        return self.request.delete_cookie("token")

    def logout(self):
        """Deletes the state depending on the implementation of this driver.

        Returns:
            bool
        """
        self.delete()
        self.request.reset_user()



"""SMTP Driver Module."""

import smtplib
import ssl
import warnings
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from ...contracts.MailContract import MailContract
from ...drivers import BaseMailDriver


class MailSmtpDriver(BaseMailDriver, MailContract):
    """Mail smtp driver."""

    def message(self):
        """Creates a message object for the underlying driver.

        Returns:
            email.mime.multipart.MIMEMultipart
        """
        message = MIMEMultipart("alternative")
        message["Subject"] = self.message_subject
        message["From"] = self.mail_from_header
        message["To"] = self.mail_to_header
        message["Reply-To"] = self.message_reply_to

        # Attach both mimetypes if they exist.
        if self.html_content:
            message.attach(MIMEText(self.html_content, "html"))

        if self.text_content:
            message.attach(MIMEText(self.text_content, "plain"))

        return message

    def send(self, message=None, message_contents=None):
        """Send the message through SMTP.

        Keyword Arguments:
            message {string} -- The HTML message to be sent to SMTP. (default: {None})

        Returns:
            None
        """
        # The old argument name was `message_contents`. users might have used this as keyword argument or as arg.
        assert (
            message is None or message_contents is None
        ), 'using deprecated  argument "message_contents" together with the new arg "message" ??'
        message_contents = message or message_contents
        if message_contents and isinstance(message_contents, str):
            warnings.warn(
                "Passing message_contents to .send() is a deprecated. Please use .text() and .html().",
                category=DeprecationWarning,
                stacklevel=2,
            )
            message = self._get_message_for_send_deprecated(message_contents)

        # The above should be removed once deprecation time period passed.
        elif not message:
            message = self.message()

        self._smtp_connect()

        if self._queue:
            from wsgi import container
            from ... import Queue

            container.make(Queue).push(
                self._send_mail,
                args=(self.mail_from_header, self.to_addresses, message),
            )
            return

        return self._send_mail(self.mail_from_header, self.to_addresses, message)

    def _smtp_connect(self):
        """Sets self.smtp to an instance of `smtplib.SMTP`
        and connects using configuration in config.DRIVERS.smtp
        Returns:
            None
        """
        config = self.config.DRIVERS["smtp"]
        if "ssl" in config and config["ssl"] is True:
            self.smtp = smtplib.SMTP_SSL(
                "{0}:{1}".format(config["host"], config["port"])
            )
        else:
            self.smtp = smtplib.SMTP("{0}:{1}".format(config["host"], config["port"]))

        # Check if TLS enabled
        if "tls" in config and config["tls"] is True:
            # Define secure TLS connection
            context = ssl.create_default_context()
            context.check_hostname = False

            # Check if correct response code for starttls is received from the server
            if self.smtp.starttls(context=context)[0] != 220:
                raise smtplib.SMTPNotSupportedError(
                    "Server is using untrusted protocol."
                )

        if config.get("login", True):
            self.smtp.login(config["username"], config["password"])

    def _send_mail(self, *args):
        """Wrapper around sending mail so it can also be used for queues."""
        mail_from_header, to_addresses, message = args
        response = self.smtp.send_message(message)
        self.smtp.quit()
        return response



from orator.migrations import Migration


class CreateFailedJobsTable(Migration):
    def up(self):
        """Run the migrations."""
        with self.schema.create("failed_jobs") as table:
            table.increments("id")
            table.string("driver")
            table.string("channel")
            table.binary("payload")
            table.timestamp("failed_at")
            table.timestamps()

    def down(self):
        """Revert the migrations."""
        self.schema.drop("failed_jobs")



"""Displays Information Command."""
import math
import os
import platform
import sys
import psutil

from cleo import Command
from tabulate import tabulate

from ..__version__ import __version__


class InfoCommand(Command):
    """
    Displays environment info for debugging.

    info
    """

    def handle(self):
        from ..cli import application

        rows = []

        rows.append(["System Information", self._get_system_info()])
        mem = math.ceil(psutil.virtual_memory().total / 1024 / 1024 / 1024.0)
        rows.append(["System Memory", str(mem) + " GB"])
        rows.append(["Python Version", self._get_python_info()])
        rows.append(["Virtual Environment", self._check_virtual_environment()])
        rows.append(["Masonite Version", __version__])
        rows.append(["Craft Version", application._version])

        if "APP_ENV" in os.environ:
            rows.append(["APP_ENV", os.environ.get("APP_ENV")])

        if "APP_DEBUG" in os.environ:
            rows.append(["APP_DEBUG", os.environ.get("APP_DEBUG")])

        self.info("")
        self.info(tabulate(rows, headers=["Environment Information", ""]))
        self.info("")

    def _get_python_info(self):
        py_version = platform.python_version()
        py_implementation = platform.python_implementation()
        return py_implementation + " " + py_version

    def _check_virtual_environment(self):
        if hasattr(sys, "real_prefix") or "VIRTUAL_ENV" in os.environ:
            return u"\u2713"  # currently running in virtual env
        return "X"

    def _get_system_info(self):
        bits, _ = platform.architecture()
        operating_system, _, _, _, arch, _ = platform.uname()

        if operating_system.lower() == "darwin":
            operating_system = "MacOS"
        return "{} {} {}".format(operating_system, arch, bits)



"""An AppProvider Service Provider."""

from ..autoload import Autoload
from ..commands import (
    AuthCommand,
    CommandCommand,
    ControllerCommand,
    DownCommand,
    InfoCommand,
    InstallCommand,
    JobCommand,
    KeyCommand,
    MailableCommand,
    MakeMigrationCommand,
    MiddlewareCommand,
    MigrateCommand,
    MigrateRefreshCommand,
    MigrateResetCommand,
    MigrateRollbackCommand,
    MigrateStatusCommand,
    ModelCommand,
    ModelDocstringCommand,
    ProviderCommand,
    PublishCommand,
    PresetCommand,
    QueueTableCommand,
    QueueWorkCommand,
    RoutesCommand,
    SeedCommand,
    SeedRunCommand,
    ServeCommand,
    TestCommand,
    TinkerCommand,
    UpCommand,
    ViewCommand,
)
from ..exception_handler import DumpHandler, ExceptionHandler
from ..helpers import config, load
from ..helpers.routes import flatten_routes
from ..hook import Hook
from ..provider import ServiceProvider
from ..request import Request
from ..response import Response
from ..routes import Route


class AppProvider(ServiceProvider):
    def register(self):
        self.app.bind("HookHandler", Hook(self.app))
        self.app.bind("WebRoutes", flatten_routes(load("routes.web.routes")))
        self.app.bind("Route", Route())
        self.app.bind("Request", Request())
        self.app.simple(Response(self.app))
        self.app.bind("Container", self.app)
        self.app.bind("ExceptionHandler", ExceptionHandler(self.app))
        self.app.bind("ExceptionDumpExceptionHandler", DumpHandler)

        self.app.bind("RouteMiddleware", config("middleware.route_middleware"))
        self.app.bind("HttpMiddleware", config("middleware.http_middleware"))
        self.app.bind("staticfiles", config("storage.staticfiles", {}))

        # Insert Commands
        self._load_commands()

        self._autoload(config("application.autoload"))

    def boot(self, request: Request, route: Route):
        self.app.bind("StatusCode", None)
        route.load_environ(self.app.make("Environ"))
        request.load_environ(self.app.make("Environ")).load_app(self.app)

    def _autoload(self, directories):
        Autoload(self.app).load(directories)

    def _load_commands(self):
        self.commands(
            AuthCommand(),
            CommandCommand(),
            ControllerCommand(),
            DownCommand(),
            InfoCommand(),
            InstallCommand(),
            JobCommand(),
            KeyCommand(),
            MailableCommand(),
            MakeMigrationCommand(),
            MiddlewareCommand(),
            MigrateCommand(),
            MigrateRefreshCommand(),
            MigrateResetCommand(),
            MigrateStatusCommand(),
            MigrateRollbackCommand(),
            ModelCommand(),
            ModelDocstringCommand(),
            PresetCommand(),
            ProviderCommand(),
            PublishCommand(),
            QueueWorkCommand(),
            QueueTableCommand(),
            ViewCommand(),
            RoutesCommand(),
            ServeCommand(),
            SeedCommand(),
            SeedRunCommand(),
            TestCommand(),
            TinkerCommand(),
            UpCommand(),
        )



"""Bootstrap Preset"""
import os
import shutil
from ..presets import Preset


class Bootstrap(Preset):
    """Configure the front-end scaffolding for the application to use Bootstrap"""

    def install(self):
        """Install the preset"""
        self.update_packages()
        self.update_sass()
        self.remove_node_modules()

    def update_package_array(self, packages={}):
        """Updates the packages array to include Bootstrap specific packages"""
        packages["bootstrap"] = "^4.0.0"
        packages["jquery"] = "^3.2"
        packages["popper.js"] = "^1.12"
        return packages

    def update_sass(self):
        """Copies Bootstrap scss files into application"""
        directory = "resources/sass"
        if not os.path.exists(os.path.realpath(directory)):
            os.makedirs(os.path.realpath(directory))
        shutil.copyfile(
            os.path.dirname(__file__) + "/bootstrap-stubs/_variables.scss",
            "resources/sass/_variables.scss",
        )
        shutil.copyfile(
            os.path.dirname(__file__) + "/bootstrap-stubs/app.scss",
            "resources/sass/app.scss",
        )



"""Upload Manager Module."""

from ..contracts import UploadManagerContract
from .Manager import Manager


class UploadManager(Manager, UploadManagerContract):
    """Manages all upload drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = "storage"
    driver_prefix = "Upload"


class Upload:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



from ..app import App
from ..request import Request
from ..response import Response


class ResponseMiddleware:
    def __init__(self, request: Request, app: App, response: Response):
        self.request = request
        self.app = app
        self.response = response

    def after(self):
        if self.request.redirect_url:
            self.response.redirect(self.request.redirect_url, status=302)
            self.request.reset_redirections()

        if self.app.has("Session") and self.request.is_status(200):
            try:
                self.app.make("Session").driver("memory").reset(flash_only=True)
            except Exception:
                pass



"""Async Driver Method."""

import inspect
import pickle
import time

import pendulum
from ...contracts import QueueContract
from ...drivers import BaseQueueDriver
from ...helpers import HasColoredCommands, parse_human_time
from ...queues import Queueable


class QueueDatabaseDriver(BaseQueueDriver, HasColoredCommands, QueueContract):
    """Queue Aysnc Driver."""

    def __init__(self):
        """Queue Async Driver.

        Arguments:
            Container {masonite.app.App} -- The application container.
        """
        pass

    def connect(self):
        return self

    def push(self, *objects, args=(), kwargs={}, **options):
        """Push objects onto the async stack.

        Arguments:
            objects {*args of objects} - This can be several objects as parameters into this method.
            options {**kwargs of options} - Additional options for async driver
        """

        from config.database import DB as schema

        callback = options.get("callback", "handle")
        wait = options.get("wait", None)
        connection = options.get("connection", None)

        if connection:
            schema = schema.connection(connection)

        if wait:
            wait = parse_human_time(wait).to_datetime_string()

        for job in objects:
            if schema.get_schema_builder().has_table("queue_jobs"):
                payload = pickle.dumps(
                    {"obj": job, "args": args, "kwargs": kwargs, "callback": callback}
                )
                schema.table("queue_jobs").insert(
                    {
                        "name": str(job),
                        "serialized": payload,
                        "created_at": pendulum.now().to_datetime_string(),
                        "attempts": 0,
                        "ran_at": None,
                        "wait_until": wait,
                    }
                )

    def consume(self, channel, **options):  # skipcq
        from config.database import DB as schema, DATABASES
        from wsgi import container

        if not channel or channel == "default":
            channel = DATABASES["default"]

        self.info(
            '[*] Waiting to process jobs from the "queue_jobs" table on the "{}" connection. To exit press CTRL + C'.format(
                channel
            )
        )
        schema = schema.connection(channel)
        while True:
            builder = schema.table("queue_jobs")
            jobs = builder.where_null("ran_at").where(schema.table("queue_jobs").where_null('wait_until').or_where('wait_until', '<=', pendulum.now().to_datetime_string())).limit(1).get()

            if not jobs.count():
                time.sleep(5)

            for job in jobs:
                builder.where("id", job["id"]).update(
                    {
                        "ran_at": pendulum.now().to_datetime_string(),
                    }
                )
                unserialized = pickle.loads(job.serialized)
                obj = unserialized["obj"]
                args = unserialized["args"]
                callback = unserialized["callback"]
                ran = job.attempts

                try:
                    try:
                        if inspect.isclass(obj):
                            obj = container.resolve(obj)

                        getattr(obj, callback)(*args)

                    except AttributeError:
                        obj(*args)

                    try:
                        # attempts = 1
                        builder.where("id", job["id"]).update(
                            {
                                "ran_at": pendulum.now().to_datetime_string(),
                                "attempts": job["attempts"] + 1,
                            }
                        )
                        self.success("[\u2713] Job Successfully Processed")
                    except UnicodeEncodeError:
                        self.success("[Y] Job Successfully Processed")
                except Exception as e:  # skipcq
                    self.danger("Job Failed: {}".format(str(e)))

                    if not obj.run_again_on_fail:
                        # ch.basic_ack(delivery_tag=method.delivery_tag)
                        builder.where("id", job["id"]).update(
                            {
                                "ran_at": pendulum.now().to_datetime_string(),
                                "failed": 1,
                                "attempts": job["attempts"] + 1,
                            }
                        )

                    if ran < obj.run_times and isinstance(obj, Queueable):
                        time.sleep(1)
                        builder.where("id", job["id"]).update(
                            {"attempts": job["attempts"] + 1}
                        )
                        continue
                    else:
                        builder.where("id", job["id"]).update(
                            {
                                "attempts": job["attempts"] + 1,
                                "ran_at": pendulum.now().to_datetime_string(),
                                "failed": 1,
                            }
                        )

                        if hasattr(obj, "failed"):
                            getattr(obj, "failed")(unserialized, str(e))

                        self.add_to_failed_queue_table(unserialized, driver="database")

            time.sleep(5)
